#include "fflas-ffpack/fflas-ffpack.h"

// See test-multifile2.C - it is a test
// to confirm that the lib is *really* header-only and full inline.

/* -*- mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
// vim:sts=4:sw=4:ts=4:et:sr:cino=>s,f0,{0,g0,(0,\:0,t0,+0,=s
